# -*- coding: utf-8 -*-
from AccessControl import getSecurityManager
from plone.keyring.interfaces import IKeyManager
from plone.protect.interfaces import IAuthenticatorView
from Products.Five import BrowserView
from zExceptions import Forbidden
from zope.component import getUtility
from zope.interface import implementer
from ZPublisher.HTTPRequest import HTTPRequest

import hmac
import six


try:
    from hashlib import sha1 as sha
except ImportError:
    import sha


ANONYMOUS_USER = "Anonymous User"


def isAnonymousUser(user):
    return user is None or user.getUserName() == ANONYMOUS_USER


def _getUserName():
    user = getSecurityManager().getUser()
    if user is None:
        return ANONYMOUS_USER
    return user.getUserName()


def _is_equal(val1, val2):
    """
    constant time comparison
    """
    if not isinstance(val1, six.string_types) or not isinstance(val2, six.string_types):
        return False
    if len(val1) != len(val2):
        return False
    result = 0
    for x, y in zip(val1, val2):
        result |= ord(x) ^ ord(y)
    return result == 0


def _getKeyring(username, manager=None):
    if manager is None:
        manager = getUtility(IKeyManager)
    if username == ANONYMOUS_USER:
        try:
            ring = manager[u'_anon']
        except KeyError:
            # no anonymous key defined.
            # XXX should we even bother allowing to verify?
            ring = manager[u'_system']
    else:
        try:
            ring = manager[u"_forms"]
        except KeyError:
            ring = manager[u'_system']
    return ring


def _verify_request(request, extra='', name='_authenticator', manager=None):
    auth = request.get(name)
    if auth is None:
        auth = request.getHeader('X-CSRF-TOKEN')
        if auth is None:
            return False
    if isinstance(auth, list):
        # in case 2 auth attributes were added to form. It can happen
        auth = auth[0]

    user = _getUserName()
    ring = _getKeyring(user, manager=manager)

    if six.PY3:
        user = user.encode('utf-8')
        extra = extra.encode('utf-8')

    for key in ring:
        if key is None:
            continue
        correct = hmac.new(key.encode('utf-8'), user + extra, sha).hexdigest()
        if _is_equal(correct, auth):
            return True

    return False

# We had to rename because previous hotfixes patched _verify
_verify = _verify_request


def createToken(extra='', manager=None):
    user = _getUserName()
    ring = _getKeyring(user, manager=manager)
    secret = ring.random()
    if six.PY3:
        secret = secret.encode('utf-8')
        user = user.encode('utf-8')
        extra = extra.encode('utf-8')
    return hmac.new(secret, user + extra, sha).hexdigest()


@implementer(IAuthenticatorView)
class AuthenticatorView(BrowserView):

    def token(self, extra=''):
        return createToken(extra)

    def authenticator(self, extra='', name='_authenticator'):
        auth = createToken(extra)
        return '<input type="hidden" name="%s" value="%s"/>' % (name, auth)

    def verify(self, extra='', name="_authenticator"):
        return _verify_request(self.request, extra=extra, name=name)


def check(request, extra='', name="_authenticator", manager=None):
    if isinstance(request, HTTPRequest):
        if not _verify_request(request, extra=extra,
                               name=name, manager=manager):
            raise Forbidden('Form authenticator is invalid.')


def CustomCheckAuthenticator(extra='', name='_authenticator'):
    def _check(request):
        return check(request, extra=extra, name=name)
    return _check


__all__ = ["AuthenticatorView", "check", "createToken",
           "CustomCheckAuthenticator"]



# -*- coding: utf-8 -*-
from AccessControl.SecurityManagement import newSecurityManager
from AccessControl.SecurityManagement import noSecurityManager
from AccessControl.User import User
from plone.keyring.interfaces import IKeyManager
from plone.keyring.keymanager import KeyManager
from unittest import TestCase
from zope.component import getGlobalSiteManager


class MockRequest(dict):

    def __init__(self, URL=None, *args, **kwargs):
        super(MockRequest, self).__init__(*args, **kwargs)
        self.environ = {}
        self.URL = URL

    def setReferer(self, url):
        self.environ['HTTP_REFERER'] = url

    def getHeader(self, name):
        return None


class KeyringTestCase(TestCase):

    key_size = 1

    def setUp(self):
        self.sm = getGlobalSiteManager()
        self.manager = KeyManager(self.key_size)
        self.sm.registerUtility(self.manager, provided=IKeyManager,
                                event=False)
        # Tests modify the user object so we better make sure it is *our*
        # user object and not the built-in Anonymous User.
        newSecurityManager(None, User('dummy', 'secret', (), ()))

    def tearDown(self):
        self.sm.unregisterUtility(self.manager, provided=IKeyManager)
        noSecurityManager()



# -*- coding: utf-8 -*-
from plone.protect.authenticator import check as CheckAuthenticator
from plone.protect.authenticator import createToken
from plone.protect.authenticator import CustomCheckAuthenticator
from plone.protect.postonly import check as PostOnly
from plone.protect.utils import protect



# -*- coding: utf-8 -*-
from plone.protect.interfaces import IConfirmView
from Products.CMFCore.utils import getToolByName
from Products.Five import BrowserView
from zExceptions import Forbidden
from zope.interface import implementer


@implementer(IConfirmView)
class ConfirmView(BrowserView):

    def __call__(self):
        urltool = getToolByName(self.context, 'portal_url')
        original_url = getattr(self.request, 'original_url', '')
        if not original_url or not urltool.isURLInPortal(original_url):
            raise Forbidden('url not in portal: {0}'.format(original_url))
        return self.index()



# -*- coding: utf-8 -*-
from plone.protect.auto import safeWrite
from zope.testing.cleanup import addCleanUp
import inspect


def wl_lockmapping(self, killinvalids=0, create=0):
    has_write_locks = hasattr(self, '_dav_writelocks')
    locks = self._old_wl_lockmapping(killinvalids=killinvalids, create=create)
    try:
        safeWrite(locks)
        if not has_write_locks and create:
            # first time writing to object, need to mark it safe
            safeWrite(self)
    except AttributeError:
        # not a persistent class, ignore
        pass
    return locks


def pluggableauth__getCSRFToken(request):
    """
    let plone.protect do it's job
    """
    return ''


def pluggableauth__checkCSRFToken(request, token='csrf_token', raises=True):
    """
    let plone.protect do it's job
    """
    pass


def marmoset_patch(func, replacement):
    source = inspect.getsource(replacement)
    exec(source, func.__globals__)
    func._old_code = func.__code__
    func.__code__ = replacement.__code__


def disable_zope_csrf_checks():
    from Products.PluggableAuthService import utils as pluggable_utils
    if hasattr(pluggable_utils, 'checkCSRFToken'):
        marmoset_patch(
            pluggable_utils.checkCSRFToken,
            pluggableauth__checkCSRFToken,
        )
    if hasattr(pluggable_utils, 'getCSRFToken'):
        marmoset_patch(
            pluggable_utils.getCSRFToken, pluggableauth__getCSRFToken)


def enable_zope_csrf_checks():
    from Products.PluggableAuthService import utils as pluggable_utils
    if hasattr(pluggable_utils, 'checkCSRFToken'):
        try:
            pluggable_utils.checkCSRFToken.__code__ = \
                pluggable_utils.checkCSRFToken._old_code
        except AttributeError:
            pass
    if hasattr(pluggable_utils, 'getCSRFToken'):
        try:
            pluggable_utils.getCSRFToken.__code__ = \
                pluggable_utils.getCSRFToken._old_code
        except AttributeError:
            pass
addCleanUp(enable_zope_csrf_checks)



# -*- coding: utf-8 -*-
from setuptools import find_packages
from setuptools import setup


version = '4.1.8'

setup(
    name='plone.protect',
    version=version,
    description="Security for browser forms",
    long_description='%s\n%s' % (
        open("README.rst").read(),
        open("CHANGES.rst").read()
    ),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Framework :: Plone",
        "Framework :: Plone :: 5.2",
        "Framework :: Plone :: 6.0",
        "Framework :: Plone :: Core",
        "Framework :: Zope :: 2",
        "Framework :: Zope :: 4",
        "Framework :: Zope :: 5",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    keywords='zope security CSRF',
    author='Plone Foundation',
    author_email='plone-developers@lists.sourceforge.net',
    url='https://github.com/plone/plone.protect',
    license='BSD',
    packages=find_packages(),
    namespace_packages=['plone', ],
    include_package_data=True,
    zip_safe=False,
    python_requires='>=2.7,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*,!=3.4.*,!=3.5.*',
    install_requires=[
        'lxml[cssselect]',
        'setuptools',
        'plone.keyring >= 3.0dev',
        'six',
        'zope.component',
        'zope.interface',
        'Zope2',
        'plone.transformchain',
        'repoze.xmliter>=0.3',
        'collective.monkeypatcher',
        'z3c.zcmlhook',
    ],
    extras_require={
        'test': [
            'plone.app.testing',
            'Products.CMFPlone'
            'zope.annotation',
        ],
    }
)
